from tensorflow.python.keras import Input, Model
from tensorflow.python.keras.datasets import mnist
from tensorflow.keras.optimizers import Adam
from tensorflow.python.keras import Sequential, Input, Model
from tensorflow.python.keras.layers import Dense, LeakyReLU, BatchNormalization, Reshape, Flatten
import numpy as np

class GNA():
    def __init__(self):
        # 定义输入图片的尺寸、通道数和噪声的维数
        self.img_rows = 28
        self.img_cols = 28
        self.channels = 1
        self.img_shape = (self.img_rows, self.img_cols, self.channels)
        self.latent_dim = 100

        optimizer = Adam(0.0002, 0.5)

        # 实例化判别器并指定训练时使用的优化器、loss等
        self.discriminator = self.build_discriminator()
        self.discriminator.compile(loss='binary_crossentropy',
                                   optimizer=optimizer,
                                   metrics=['accuracy'])

        # Build the generator
        self.generator = self.build_generator()

        # 为生成器设置一个输入层Input，形状为100维
        z = Input(shape=(self.latent_dim,))
        img = self.generator(z)

        '''
        https://blog.csdn.net/qq_38669138/article/details/109029782
        说明：这里设置不训练判别器并不是判别器就一直不会训练
        在train()中还是会通过train_on_batch的方式去训练判别器的
        原因：
            当discriminator被compile之后，即使设置了discriminator.trainable=False，
        该discriminator仍然可以通过train_on_batch的方式被训练；
    
            但是如果discriminator在被compile之前就把训练状态设置为False，
        那么即使是使用discriminator.train_on_batch的方式也不能训练该判别器。
        '''
        self.discriminator.trainable = False

        # 将生成器生成的图片输入判别器
        validity = self.discriminator(img)
        '''
        使用Model构建一个网络
        该网络由一个生成器层和一个判别器层组成
        通过训练该网络，让生成器尽可能骗过判别器，也就是说，该网络的作用是训练生成器
        由于之前的设置，该网络训练过程中，其判别器层不训练
        '''
        self.combined = Model(z, validity)
        self.combined.compile(loss='binary_crossentropy', optimizer=optimizer)

    def build_generator(self):
        # 使用Sequential方式构建生成器

        model = Sequential()

        # 输入为100维的随机噪声数据，经过第一层全连接层后输出为256维
        model.add(Dense(256, input_dim=self.latent_dim))
        model.add(LeakyReLU(alpha=0.2))
        model.add(BatchNormalization(momentum=0.8))

        model.add(Dense(512))  # 经过第二层全连接层，输出为512维
        model.add(LeakyReLU(alpha=0.2))
        model.add(BatchNormalization(momentum=0.8))

        model.add(Dense(1024))  # 经过第三层全连接层，输出为1024维
        model.add(LeakyReLU(alpha=0.2))
        model.add(BatchNormalization(momentum=0.8))

        # 经过第四层全连接层，输出为设置的图片的尺寸（这里是28*28，也就是784维）
        model.add(Dense(np.prod(self.img_shape), activation='tanh'))
        model.add(Reshape(self.img_shape))  # Reshape成1*28*28

        model.summary()  # 打印网络结构

        # 下面两句就是给定输入，获取一个输出
        # 有点类似pytorch模型类中forward函数的意思
        noise = Input(shape=(self.latent_dim,))
        img = model(noise)
        return Model(noise, img)  # 返回构造好的模型

    def build_discriminator(self):
        # 使用Sequential方式构建判别器
        model = Sequential()

        # 将输入的图片展平并输入全连接层
        model.add(Flatten(input_shape=self.img_shape))
        model.add(Dense(512))
        model.add(LeakyReLU(alpha=0.2))

        model.add(Dense(256))
        model.add(LeakyReLU(alpha=0.2))

        model.add(Dense(1, activation='sigmoid'))  # 最后通过维度是1的全连接层
        model.summary()

        img = Input(shape=self.img_shape)
        validity = model(img)

        return Model(img, validity)

    def train(self, epochs=30, batch_size=128, sample_interval=50):
        # 载入数据集
        (X_train, _), (_, _) = mnist.load_data()

        # 将数据调整到-1~1范围上
        X_train = X_train / 127.5 - 1.
        X_train = np.expand_dims(X_train, axis=3)  # 增加X_train的维度

        # 创建标签(判断为真/假,1/0)
        valid = np.ones((batch_size, 1))
        fake = np.zeros((batch_size, 1))

        for epoch in range(epochs):

            # ---------------------
            #  Train Discriminator
            # ---------------------
            # 根据索引，随机从数据集中选择图片数据
            idx = np.random.randint(0, X_train.shape[0], batch_size)
            imgs = X_train[idx]

            noise = np.random.normal(0, 1, (batch_size, self.latent_dim))

            # 使用Sequential建立的模型可以通过predict方法返回输出值
            gen_imgs = self.generator.predict(noise)

            # 输入真实图片与生成图片训练单独判别器
            d_loss_real = self.discriminator.train_on_batch(imgs, valid)
            d_loss_fake = self.discriminator.train_on_batch(gen_imgs, fake)
            d_loss = 0.5 * np.add(d_loss_real, d_loss_fake)

            # ---------------------
            #  Train Generator
            # ---------------------
            # 产生正态分布的随机噪声
            noise = np.random.normal(0, 1, (batch_size, self.latent_dim))

            # 训练combined网络，注意，其中的判别器不训练
            g_loss = self.combined.train_on_batch(noise, valid)

            print("%d [D loss: %f, acc.: %.2f%%] [G loss: %f]" % (epoch, d_loss[0], 100 * d_loss[1], g_loss))

            if epoch % sample_interval == 0:
                self.sample_images(epoch)